#!/usr/bin/perl
$FORMAT_TOP_NAME = "ddd";
use Carp;
use English;
use Getopt::Std;
 
# This program is the code for STEMRUST_G.pl, a simulation model for stem rust epidemics in perennial ryegrass seed crops
# Produced by Bill Pfender as an employee of USDA Agricultural Research Service, and Dan Upper. It is in the public domain.
# This program should be considered to be research code, not a user-friendly finished program.
# Other files have documentation and ancillary code ("utilities_pruned.pl") needed for some of the calculations  

BEGIN {
    $islinux = ($^O eq "linux")? 1 : 0;
    $iswindows = 1-$islinux;
}
     require "utilities_pruned.pl";  # can use "util-rust_v2.pl",it includes everything that "utilities_pruned.pl" has   

my $today;

# list of subroutines
sub push_group( $\% );   # print a group of nine numbers given a hash
sub new_array ( $$ );    # (value, length) initialize an array; returns
      # an array reference.  Initialization isn't normally necessary, but I
      # wanted to deal with fungicide effects a week in the future.
sub add ( @@ );          # (@a,@b) both arrays $result[i] = $a[i]+$b[i]
sub mult ( @@ );         # (@a,@b) both arrays $result[i] = $a[i]*$b[i]
sub dot ( @@ );
sub power ( $$ );
sub make_fungicide_factor ( $$$ ) ;  # creates an array of survival rates
sub apply_kickback ( $$ ); # reduces cohort size as appropriate
sub plant_area ( $ );    # (day) area in square cm
sub thresholdTriaz ( $ );     # threshold for making spray decision about triazole fungicide
sub thresholdStrob ( $ );     # threshold for making spray decision about strobilurin fungicide
sub sheath_length ( $$ );  # (sheath number, day) length in cm
sub stem_length ( $$ );  # (stem number, day) (flower = stem0) length in cm
sub barestem_length ( $$ ); # length of sub-node stem (non-sheath part of 'stem')
sub new_cohort ( $$$;$$ );# $from, $on, $size, $srcnum, $snum. Create a new cohort of infections
sub dumpcohort ( $% );   # args are (title and hash reference). this collects data from cohort hashes to print out
sub age_cohort ( $ );    # move latent to erumpent as per LUP accumulation
sub expose_cohort ( $ ); # move covered to exposed as per sheath+stem growth;
sub stem_infs( $$ );     # component of code to create contact-spread infections
sub inoculum();          # compute inoculum
sub new_primaries();     # return list of new aerial-spread-infection cohorts; primaries aka aerial-spread
sub new_secondaries();   # return list of new contact-spread-infection cohorts; secondaries aka contact-spread
sub barestem_portion( $ ); #$shnum; fraction of plant area in barestem shnum
sub plant_area_decrease();
sub simulation( $$$ );   # the actual simulation -- init and loop over days
sub pop_summary ($ );   
sub accum_stats ( @ );   # accumulate vis, all-infs, alive and dead-or-alive
sub count_sec_infs ( );  # this is only used for output, not part of model operation
sub count_sec_vis ( );   # this is only used for output, not part of model operation

# following is a list of switches that can be used in the command line to alter settings; then description of the options 
# process flags on command line
($opt_a,$opt_b,$opt_c,$opt_d,$opt_e,$opt_f,$opt_h,$opt_i, $opt_k,$opt_l, $opt_n,
 $opt_p,$opt_q,$opt_s,$opt_t, $opt_u,$opt_w, $opt_x, $opt_y, $opt_z)=
    (undef,undef,undef,undef, undef, undef,undef,undef,undef, undef, undef,
     undef,undef,undef,undef, undef, undef, undef, undef, undef);
getopts('a:b:c:d:e:f:hi:k:l:np:q:s:t:u:wxy:z:');  # a colon means there is an argument for the option

if ($opt_h) {
    print <<"EOF";
usage: $0 [options] input_file
where options are:
   -a number	 sets $stem_infs_pcmL for testing (default is now 15)
   -b filename   output of cohort hash file
   -c rain cnst  sets value of $prior_rain_constant for testing
   -d obs day    day observations were collected   (no default)
   -e end day    day to end simulation (default -- when data is gone)
   -f filename   send results to a text file
   -h            help -- print this message
   -i inoculum   initial "seed" pustules on rust start day. (default 0.01)
   -k inf k      sets the value of $k_infection, for model testing
   -l pust life  sets duration of pustule life, counting from Lup at initiation 
   -p pustules   number of pustules / sq ft seen on obs day (no default)
   -q day1,day2  days quadris was sprayed (comma-sep list, default is empty)
   -s day        sheath 1 observed date (mm/dd). Used to determine stem start date for model.
   -t day1,day2  days tilt was sprayed (comma-sep list, default is empty)
   -u day1,day2  days quilt was sprayed (comma-sep list, default is empty)
   -w            print standard error output for use in plotting
   -x            text output -- include page headers
   -y            debug output file
   -z            to include recently dead pustules in visible disease, set this to lup since infection	

Solver mode: if either -d or -p is used, $0 is in "solver mode", which
requires both -d and -p and prohibits -i and -m.  This mode runs the
simulation repeatedly to find an inoculum level which
produces

Output:
   Five columns output are printed per day of year:
   1) 'day'  = date in m-dd format
   2) 'cumu vis' = vis_live = cumulative visible live pustules ( eep + ees) 
   3) 'cumu infs'= infs_live= cumulative infections (eep + ees + lep + les + lcs + ecs};
   4) 'inf factor' = infection factor for the day
   5) 'lup' = latent unit period. The basic unit of physiological time.
   6) 'Sheath_length(2)' = length of sheath 2
   7) 'Sheath_length(1)' = length of sheath 1 (flag sheath)
   8)  'stem_length(0)' = length of inflorescence

EOF
    exit(0);
}

$stderrout = $opt_w;  # output for plot script

@tilt_dates = split ',', $opt_t if $opt_t;
@tilt_days=();
for $date (@tilt_dates) {
    push @tilt_days, make_doy($date);
}
 
@quadris_dates = split ',', $opt_q if $opt_q;
@quadris_days=();
for $date (@quadris_dates) {
    push @quadris_days, make_doy($date);
}

@quilt_dates = split ',', $opt_u if $opt_u;
@quilt_days=();
for $date (@quilt_dates) {
    push @quilt_days, make_doy($date);
} 

# release of inoculum, in percent of total per day after eruption;
@inoc_schedule = (0.001, 0.005, 0.020, 0.056, 0.126, 0.120, 0.107, 0.093,
                  0.079, 0.068, 0.057, 0.049, 0.041, 0.035, 0.0429, 0.025,
                  0.021, 0.017, 0.015, 0.012, 0.010, 0.009, 0.007);

# sheath length model parameters. Sheath 1 is the sheath around the flower stem.
# length has form $a/(1+exp($b-LUP)/$c)) - d(a). It is a sigmoid, translated by
# $b and stretched in time by $c and in height by $a. Indices are sheath numbers;
# there is no sheath 0. The term (d*a) truncates to account for the part that is enclosed within next-older sheath.

# parameters for sheath length calculations
@sheath_a = (0, 17.30, 14.30, 11.00, 8.90);
@sheath_b = (0,  4.20,  3.30,  2.70, 2.00);
@sheath_c = (0,  0.38,   0.45,  0.55, 0.60); 
@sheath_d = (0,  0.30,  0.40,  0.30,  0.50);

# stem length has the same form as sheath length, but different parameters.
@stem_a = (37.5, 21.6, 18.3, 13.6, 9.66);
@stem_b = ( 5.65, 4.6,  3.65, 2.74, 2.00);
@stem_c = ( 0.72, 0.60, 0.60, 0.54, 0.55);
@stem_d = ( 0.13, 0.21, 0.32, 0.34, 0.50);

$tillers = 325; #tillers per ft row, final number usu reached in mid-May. Actual seasonal change too small to be worth accounting for

# must have $stemstart >= $ruststart
$ruststart = datetodoy(3,1);     # rust model begins this day -- fixed

# set s value ($shth1date) in command line, or use a default value, as plant growth biofix (date flag sheath collar appears)
$shth1date = $opt_s ? $opt_s : datetodoy(5,6); 

# set kMAX value in command line, or use a default value   
 $kMAX = $opt_k ? $opt_k : 0.2;

# set a value ($stem_infs_pcmL) in command line, or use default value, for contact-spread infections per cm length of recipient stem  
$stem_infs_pcmL = $opt_a ? $opt_a : 15;  

# set prior_rain_constant value in command line, or use a default value
$prior_rain_constant = $opt_c ? $opt_c : 50;  #larger value = less effect of rain

$initial_pustules = $opt_i ? $opt_i : 0.01;     # number of "seed" pustules
#              which erupt on day $ruststart -- source of all disease.
#              if using solver, this is ignored

$end_day = $opt_e ? $opt_e : datetodoy(7, 25);  # stop simulation now  
               # altered later if data stops short

# set duration of pustule life (from initation), or use default of 2.5
$pust_life = $opt_l ? $opt_l : 2.5 ;

# set post-dying visibility of pustules in the command line, or use default value 
# of 2.5 (invisible as soon as dead of old age). Larger value keeps pustules visible (thus part of 'visible disease') after they die 
# z=2.5 keeps fungicide-killed pustules visible for normal life-span instead of making them invisible immediately; 
# a smaller value for z would take fungicide-killed pustules out of "vis dis" sooner
$vis_duration = $opt_z ? $opt_z : 2.5 ; 

$daymax = $end_day+21;        # add margin in arrays

$pustules_per_cm2 = 20;       # max density of pustules (ie available sites)

$MIN_COHORT_SIZE = 1e-6;      # don't create infinitesimal cohorts

$AUDPC = 0;                   # declare AUDPC, area under disease progress 
                              #curve (uses live + recently-dead)

@area_oth_stored = (0);       # storage variable for parts area "9" ie "other", 
                              # for use in area decrease function


######### instructions for output to screen  #####
##        NOTE that $looks_live is same as $vis_live most of the time, but adds the recently-killed
##          pustules for the 5 days starting on fungicide application day (because they look alive)

# following is the currently-used screen output
format sim_out =
   @<<<<   @####.#   @####.#    @###    @##.##   @##.#     @##.#      @##.#    @######   @######                  
   "$mm-$dd", $looks_live, $infs_live, $inf_factor, $lup-$lup[60], $dsp_sl2, $dsp_sl1, $dsp_st0, $test_threshTriaz, $test_threshStrob
# following is to output $vis_live instead of $looks_alive (no adjustment for live-looking but dead pustules after fungicide)
#   @<<<<   @######   @######.#    @###    @##.##   @##.#     @##.#      @##.#    @######   @######   
#   "$mm-$dd", $vis_live, $infs_live, $inf_factor, $lup-$lup[60], $dsp_sl2, $dsp_sl1, $dsp_st0, $test_threshTriaz, $test_threshStrob
# following is for screen output to see the replacement of vis_live with vis_AandRecD for 5 days post-fungicide
#   @<<<<   @######   @######    @#####    @####.#   @###     @##.##                   
#   "$mm-$dd", $vis_live, $vis_AandRecD, $looks_live, $infs_live, $inf_factor, $lup-$lup[60]
.
format sim_out_TOP =
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$header_string
.
format null =
.
$header_string = ' #date looks_live cumu_infs  inf_factor  lup    lngthsh2  lngthsh1  lngthst0  DMI_Test Strob_Test';
# headers for corresponding alternative outputs coded above
# $header_string = ' #date     cumu_vis cumu_infs  inf_factor  lup    lngthsh2  lngthsh1  lngthst0  DMI_Test Strob_Test';
# $header_string = ' #date     vis    vis_AandRecD  looks_live  all infs inf_factor  lup'; 
######### end of instructions for output to screen  #####

($persistence, $kb_events) = 
    make_fungicide_factor( \@quadris_days, \@quilt_days, \@tilt_days);

sub power ( $$ ) {
    my ($x, $y) = @_;
    return exp( $y*log($x) );
}

# set up arrays to read
@rain = @{ new_array( 0, $daymax)};
@inf_factor = @{ new_array( 0, $daymax)}; # infection value
@lup = @{ new_array( 0, $daymax)};
@days = @{ new_array( -1, $daymax)};

$id = undef; # warning silencer

$firsttime = 1;
# read and process data from input weather file
while(<>) {
    chomp;
    s/\#.*$//;
    s/\s//g;
    next if /^\s*$/;

    ($id, $year, $day, $time, $rain, $infval, $lup) = split ',';
#    print STDERR "id $id\tyear $year\tday $day\ttime $time\train"
#      ," $rain\tinf $infval\tlup $lup\n";

    $rain[$day]   = $rain;
    $lup[$day]    = $lup;


    $inf_factor[$day] = (power(10, $infval)-1);  # makes inf_factor (range 0 -1000), from "Infval" of ril file

    if ($day>60 and $day<200 and $days[$day-1]<0) {
        warn "missing data during growing season: day ".($day-1);
        # for now just warn; add work-arounds later if needed
    }
    if ($days[$day-1]<0 and not $firsttime ) {
        # $day-1 missing from input; interpolate it naively
        my ($tmp, $prev);
        $prev = $day-1;
        while($days[--$prev]<0){}  # how many days are missing?
        my $rain_interp = 0.5*($rain[$prev] + $rain[$day]);  # arithm. mean
        my $lupperday = ($lup[$day]-$lup[$prev])/($day-$prev);  # linear
        my $inf_interp = sqrt($inf_factor[$prev]*$inf_factor[$day]); #geo mean
        for($tmp=$prev+1; $tmp<$day; $tmp++) {
            $rain[$tmp] = $rain_interp;
            $lup[$tmp] = $lup[$prev] + ($tmp-$prev)*$lupperday;
            $inf_factor[$tmp] = $inf_interp;
        }
    }

    $days[$day]=0;
    undef $firsttime;
}


$end_day=$day unless $end_day < $day;
$daymax = $end_day+21;

# stem model LUP starts now
my $stemstart = &datetodoy(3,1);     # Default
my $stem_start_lup = 0;              # Default is 0
my $sheath1_start_lup = 3.9;         # Default
my $lup_btwn_stemstrt_sh1strt = 3.9;

# biofix for plant growth
my $sheath1_start_doy = &make_doy($shth1date); #sheath1 start, as day number, set by opt_s
if ($sheath1_start_doy > $end_day) {           # if opt_s is later than obs day (ie, if predicting the opt_s date)
   $sheath1_start_lup = $lup[$end_day] + ($sheath1_start_doy - $end_day)*0.084; 
} else {
       
   $sheath1_start_lup = $lup[$sheath1_start_doy];
}

$stem_start_lup = $sheath1_start_lup - $lup_btwn_stemstrt_sh1strt;

#next, replace raw lup data of ril with the lup offset to equal 0 at stemstart;

for($doy = 0; $doy <= $#lup; $doy++){
    $lup[$doy] -= $stem_start_lup if defined $lup[$doy];
}
# from here on, $lup will refer to the offset lup values, not 
# the original values in ril file

# There's no point in going beyond the days we have data for
# (unless it's to extrapolate for one LUP, or add forecast?? ....)
#$end_day=$day unless $end_day < $day;
#$daymax = $end_day+21;

#------------------------solver starts here -------------

if ($opt_d and $opt_p) { # using solver
#    die "options -d + -p are incompatible with -i and -m" if $opt_i or $opt_m;
    my ($obs_day, $obs_count) = (make_doy($opt_d), $opt_p);
    print STDERR "observation day $obs_day\n";
    print STDERR "end day         $end_day\n";
    die "can't run farther than day $end_day" if $end_day < $obs_day;
#    my $tmpfile = $iswindows ? "tmp\\$$" : "/tmp/$$"; #old version - requires existence of "tmp" subdirectory
    my $tmpfile = "solver_log_discard_$$";


    # initial values for solver drawn from 2003 field data
    # Multiply pustule count on obs. date by factor get initial value.
    # Obs. date    day       factor
    # <new year>     0        1.0     <- addition to trap errors
    # April 7       98        1.0
    # April 14     105        0.1
    # April 21     112        0.08
    # April 28     119        0.06
    # May 5        126        0.01
    # May 12       133        0.005
    # May 19       140        0.001
    my @iv_days = (0, 98,105,112,119,126,133,140);
    my @iv_factors = (1.0, 1.0,0.1,0.08,0.06,0.01,0.005,0.001);

    my $idx =$#iv_days;
    while($iv_days[$idx] > $obs_day) { --$idx }
    my $x1 = log($iv_factors[$idx]*$obs_count);

    my $ytol = log(1.05); # error in y must be < 20%, no 5%

    # We need an initial value such that: simulated disease == $obs_count.
    # We'll solve equivalent eq'n: 0==log(simulated disease)+$offset, where
    my $offset = -log($obs_count);
    #print STDERR "x1 $x1  idx: $idx offset: $offset\n";
       
    my $sim_disease = simulation(exp($x1), $tmpfile, $obs_day);
    my $y1 = log($sim_disease + 1e-6) + $offset;
    printf STDERR "1st: %5.3f %5.3f\n", $x1, $y1;   
    # done if (abs($y1) < $ytol);

    # correct for how much we're off by, but overshoot so we're (almost)
    # sure to bracket the root. [slope of y vs x is about 1, because
    # exp(y)==simulated disease is approx linear in initial disease==exp(x)]
    # should check bracketing
    my $x2 = $x1 - 1*$y1;  
    $sim_disease = simulation(exp($x2), $tmpfile, $obs_day);
    my $y2 = log($sim_disease + 1e-6) + $offset;
    printf STDERR "2nd: %5.3f %5.3f\n", $x2, $y2;   
    # done if (abs($y2) < $ytol);
    
    while(abs($y2) >= $ytol) { 
        # interpolate (or perhaps extrapolate) to find the point
        # where the line from (x1,y1) to (x2,y2) crosses the x axis.
        my $x_tmp = $x1 + ($x2-$x1)*(-$y1/($y2-$y1));
        ($x1, $y1) = ($x2, $y2);  # discard oldest pair -- should check order
        $sim_disease = simulation(exp($x_tmp), $tmpfile, $obs_day);
        ($x2, $y2) = ($x_tmp, log($sim_disease + 1e-6) + $offset);
        printf STDERR "3nd: %5.3f %5.3f\n", $x2, $y2;   
    }

    $initial_pustules = exp($x2);

    # we have now found the value we need.

} elsif ($opt_d or $opt_p) {
    # args don't make sense
    if ($opt_d) {
        die "You observed on day $opt_d,".
            "but how many pustules/sq.ft. did you see?";
    } else {
        die "You observed $opt_p pustules/sq.ft.,".
            "but when did you see them?";
    }
}
#-----------------------end solver ----------------

# --------------------prepare output files ---------------------

# output b is for viewing hash of cohort values, if desired
# the following lines are to collect cohorthash keys for print header and to run dumpcohort sub   
@cohort_keys = qw(infday eruday lup from on snum srcnum le lc ee ec eei eci de dc init_stem_len init_sheath_len);
unshift @cohort_keys, "serial no";
push @cohort_keys, "exp frac";
push @cohort_keys, "fungspor";

if ($opt_b)
{
    open COHORTcsv, ">$opt_b" or die "can't open file $opt_b";
    print COHORTcsv "title,", join(",",@cohort_keys), "\n"; 
}
# end of output b preparation

# output y is the usual output, for constructing disease progress graphs and graphs of many intermediate processes
if ($opt_y)
{
    open DEBUGFILE, ">$opt_y" or die "can't open file $opt_y";
    my @opty_out = ();
    push @opty_out, "day";            # day
    push @opty_out, "rain";           # rain in previous 24 hr
    push @opty_out, "inf_factor";     # inf factor calculated from infVal
    push @opty_out, "cumu_LUP";       # cumulative heat units (LUP)
    push @opty_out, "inoculum";       # spores released from all pustules yesterday
    push @opty_out, "visible_disease";# $looks_live, ie "vis_live", plus fungicide-killed for 5 days after application
    push @opty_out, "vis_live";       # erumpent, still-alive, pustules (omits those killed by fungicide but still appearing normal)
    push @opty_out, "vis_AandRecD";   # "vis_live" plus recently died (fungicide or old age)
    push @opty_out, "latent+visible";# "vis_live" plus latent
 #   push @opty_out, "infs_AandRecD";  # "vis_AandRecD" plus latent
    push @opty_out, "thresholdTriaz"; # action threshold for triazole, determined empirically by modeling Pacific NW US data
    push @opty_out, "thresholdStrob"; # action threshold for strobilurin, determined empirically by modeling Pacific NW US data

    # susceptibility factor for plant parts, based on their ages
    foreach (qw(stem0 sheath1 sheath2 sheath3 sheath4 other)) {
	push @opty_out, "susc $_";
    }

    # plant parts area
    foreach (@plant_parts) {
	push @opty_out, "parts_area $_";
    }

    # available sites on each class of plant parts
    foreach (@plant_parts) {
	push @opty_out, "avail $_";
    }

    # filled sites on each class of plant parts
    foreach (@plant_parts) {
	push @opty_out, "filled sites $_";
    }

    # new "primary" (aerial-spread) infections on each class of plant parts
    foreach (@plant_parts) {
	push @opty_out, "new_pri $_";
    }

    # new "secondary" (contact-spread) infections on each class 
    foreach (1..4) {
	push @opty_out, "src_infs $_";
    }

    # total "secondary" (contact-spread) infections on each class
    foreach (0..4) {
	push @opty_out, "sec_infs $_";
    }

    # barestem length, ie below the node but above next-lower sheath
    foreach (1..4) {
	push @opty_out, "barestem length $_";
    }

    # stem lengths 
    foreach (0..4) {
	push @opty_out, "stem length $_";
    }

    # sheath lengths
    foreach (1..4) {
	push @opty_out, "sheath length $_";
    }

    # total visible , live pustules that originated by contact-spread
    foreach (0..3) {
	push @opty_out, "sec_vis $_";
    }

    print DEBUGFILE join(',',@opty_out), "\n";
}
# this is the end of output y preparation


# ---------------- run the final simulation run--------------------- 
# $initial_pustules has the amount which we would have started with 

simulation ($initial_pustules, $opt_f, 0);

close COHORTcsv if $opt_b;

print STDERR "========= $0 end =========\n";
close DEBUGFILE if $opt_y;
exit 0;

# =====subroutines======================================================

#---------------- the simulation per se is the subroutine that starts here ---------
sub simulation( $$$ ) {
    my ($initial, $outfile, $obs_day)  = @_;
    my $retval = -999;
    my $been_here_before=undef;
    if($obs_day == 0)
    {
        printf STDERR "Running results\n";
    }
    else
    {
        printf STDERR "Backtracking for %d-%d\n", make_date($obs_day);
    }
    printf STDERR "Simulation is using %5.3f for initial_pustules on %d-%d\n", $initial, make_date($ruststart);

    set_output( $outfile );
    $inoculum = new_array( 0, $daymax);
    @dead_cohorts = ();  # just the dead ones
    @cohorts = ();       # just the living ones

    $seed_cohort = {
        'de'=>0,
        'dc'=>0,
        'ec' => 0,
        'ee'=>$initial,
        'from'=>'primary',
        'infday' => __UNDEF,
        'eruday' => $ruststart,
        'lc' => 0,
        'le' => 0,
        'eci' => 0,
        'eei' => 0,
        'lup' => $lup[$ruststart]-1,    # start the seed cohort on day=ruststart, it began 1.0 LUP earlier
        'on' => 'other',      
        'serial no' => 1,               #each cohort has a unique serial number
        'srcnum' => 0,
        'exp frac' => 1,
        'fungspor' => 1
    };
    push @cohorts, $seed_cohort;

##### ----------------- START MAIN LOOP ---------------#####

        #------------bookkeeping day designations -------------

    for($day = $ruststart; $day<= $end_day; $day++) {
        # day-of-year copies under convenient names
        ($mm, $dd) = doytodate($day);  # this subroutine is called from an ancillary file
        $today = $day;
        $yesterday = $day-1;
	$lup = $lup[$day];  
 
        #$lup_ago (a day number, not an lup value) is the most recent day that is at least one LUP ago.
        for($lup_ago=$yesterday; $lup< ($lup[$lup_ago] + 1); $lup_ago--) { 
            if ($lup_ago<1) {  # if the record doesn't go back that far
                undef $lup_ago;
                last;
            }
        }

      #------------ plant area calcs, and setting threshold --------------
	$thresholdTriaz = thresholdTriaz($today);
	$thresholdStrob = thresholdStrob($today); 
        $plant_area = plant_area($today);
        plant_area_decrease();


        #------------ do the "secondary" (contact-spread) infs ----------
        # new secondary cohorts, one for each stem or sheath which underlies
        # another sheath.
        my @sp = new_secondaries();
        push @cohorts, @sp if $#sp > -1;


        # -----update cohorts' life stages: expose, age and move-----
        @cohorts_tmp = ();

        foreach $cohort (@cohorts) {
            #dumphash("$count pre-aging", $cohort);
            die "undefined cohort" unless defined $cohort;
            # move population through life stages
            expose_cohort($cohort);
            my ($alive) = age_cohort( $cohort );
            if ( $alive ) {   # bookkeeping to track live and dead cohorts
                push @cohorts_tmp, $cohort;
            } else {
                push @dead_cohorts, $cohort;
            }
        }
        @cohorts = @cohorts_tmp;  #last step in updating cohorts


        #------------ "primary" (aerial-spread) infections ----------------
        $inf_factor = $inf_factor[$today];
        my @np = &new_primaries();
        push @cohorts, @np if $#np > -1;


        #--------------- apply kickback ---------------
        # kb_events->{$day} is the NAME of a fungicide applied that day
        apply_kickback($kb_events->{$day}, \@cohorts)
          if exists $kb_events->{$day}; #if there is a fungicide applied on that day

        #diagnostic (debug) output for kickback------------
    	if ($stderrout) {
    	    if (exists $kb_events->{$day}) {
    		printf STDERR " %3s\n", $kb_events->{$day}
    	    } else {
    		printf STDERR " N \n";
    	    }
    	}

     #----------------- accumulate stats for graphing ----------------
        ($vis_live, $vis_AandRecD, $infs_live, $infs_AandRecD) = 
            accum_stats(@cohorts, @dead_cohorts);

      @sec_infs = &count_sec_infs();

      @sec_vis = &count_sec_vis();

#     # print output from following lines if want to calculate whether total disease (incl rec dead) 
#         is above or below thresholds
      $test_threshTriaz = $thresholdTriaz - $infs_AandRecD;
      $test_threshStrob = $thresholdStrob - $infs_AandRecD;

      $AUDPC = AUDPC($today);

        # ------ needed for plant-growth output to webpage ------------
            ##  add new variables, if wanted, for write statements for sim-out here
	    $dsp_sl2 = &sheath_length(2,$day);
	    $dsp_sl1 = &sheath_length(1,$day);
	    $dsp_st0 = &stem_length(0,$day);


        # ------------------------main (opt_y) output------------
	    if ($opt_y) {
		my @opty_out = ();
		push @opty_out, $day;      
          	push @opty_out, $rain[$day];
          	push @opty_out, $inf_factor[$day];
          	push @opty_out, $lup[$day];
                push @opty_out, inoculum();
                push @opty_out, $looks_live;
                push @opty_out, $vis_live;
                push @opty_out, $vis_AandRecD;
                push @opty_out, $infs_live;
 #               push @opty_out, $infs_AandRecD;
                push @opty_out, $thresholdTriaz;
                push @opty_out, $thresholdStrob;

		# suscept
		foreach (qw(stem0 sheath1 sheath2 sheath3 sheath4 other)) {
		    push @opty_out, $opty_susc->{$_};
		}

		# plant parts area
		foreach (@plant_parts) {
		    push @opty_out, $opty_parts_area->{$_};
		}

		# available sites
		foreach (@plant_parts) {
		    push @opty_out, $opty_avail->{$_};
		}

		# filled sites
		foreach (@plant_parts) {
		    push @opty_out, $opty_filled->{$_};
		}

		# new primary infections
		foreach (@plant_parts) {
		    push @opty_out, $opty_new_infs->{$_};
		}

		# new secondary infections 
		foreach (1..4) {
		    push @opty_out,  $opty_src_infs->[$_];
		}

		# total secondary infections, currently does not include erumpent covered (ec)
		foreach (0..4) {
		    push @opty_out, $sec_infs[$_];
		}

		# barestem length
		foreach (1..4) {
		    push @opty_out, barestem_length($_, $today);
		}

		# stem length
		foreach (0..4) {
		    push @opty_out, stem_length($_, $today);
		}

		# sheath length
		foreach (1..4) {
		    push @opty_out, sheath_length($_, $today);
		}

		# visible, live secondaries
		foreach (0..3) {
		    push @opty_out, $sec_vis[$_];
		}


		print DEBUGFILE join(',',@opty_out), "\n";
	    } # close opt_y
        # ------------------------end opt_y---------------------------


        # -----------------to write stuff to the screen --------------
	    write;  #  writes the standard screen output


        # ------------------------ opt_b output (dumphash of cohorts) ---------
            foreach $chtdata (@cohorts, @dead_cohorts){
               dumpcohort ($today, $chtdata);
            }

    # ------ recording the value that will be used by solver (in retval) ---- 
            if ($day == $obs_day) 
            {

               printf STDERR "Simulation will predict %5.3f observed on %d-%d.\n", $vis_AandRecD, make_date($obs_day);

#     choose (un-comment) one of following lines for scouting value, in order to include (or not) dead pustules in 'visible disease'
#         But NOTE that current default settings of opt_l and opt_z eliminate the difference between $vis_live and $vis_AandRecD with
#              to dying of old age (though not with respect to dying from a fungicide application). 
 
               $retval = $vis_AandRecD;  # this one includes recently-dead (cf values of opt_z and opt_l) in observed disease
#               $retval = $vis_live;     # this one omits recently dead (whether by old-age or fungicide) from observed disease
            }

    }

    stop_output();
    return $retval;
}
# ----------------------- end of sub simulation () ----------------------------

#######------------------ other subroutines follow ----------------------------

# to produce opt_b output, the cohort hashes (alive and dead)
sub dumpcohort ($%) {  
    my ($title, $aa) = @_;
    return unless $opt_b;
    print COHORTcsv "$title,";
    #print "[", scalar(keys %$aa), "]";
    foreach (@cohort_keys) {
	print COHORTcsv defined($aa->{$_}) ? $aa->{$_} : "__UNDEF", ",";
    }
    print COHORTcsv "\n";
}

sub accum_stats ( @ ) {
    my @ch = @_;
    my ($vis_live, $vis_AandRecD, $infs_live, $infs_AandRecD) = (0,0,0,0);
    for $ch (@ch) {
       my $ee = $ch ->{'ee'};
       my $rec_dead =0;
       # following lines to include recently-dead pustules defined by opt_z value of $vis_duration
       if ($lup[$day]< $ch->{'lup'} + $vis_duration){   
          $rec_dead = $ch->{'de'};
       }

       #the pustules that will eventually become exposed because their plant-part substrate will emerge from within an older sheath 
       my $futexp = (1 - $ch->{'exp frac'}) * ($ch->{'ec'} + $ch->{'lc'}); 
       my $latent = $ch->{'le'};

       $vis_live += $ee;
       $vis_AandRecD += $ee + $rec_dead;                       #alive plus recently dead 
       $infs_live += $ee + $futexp + $latent;
       $infs_AandRecD += $ee + $futexp + $latent + $rec_dead;  #include contact-spread infs that will be exposed 

##### following block creates output to add recently-dead to vis for 5 days after fungicide application 

       my $d, $post_spray_flag=0;
       for $d (@quadris_days, @quilt_days, @tilt_days) {
           if ($day >= $d and $day <= $d+4) { $post_spray_flag = 1 }
       }

       $looks_live = $post_spray_flag ? $vis_AandRecD : $vis_live;  #new variable for output

    }
#     return ($vis_live, $vis_AandRecD, $infs_live, $infs_AandRecD);
     return ($vis_live, $vis_AandRecD, $infs_live, $infs_AandRecD, $looks_live);
}      


# initialize an array with elements 0 .. $max set to $val
sub new_array ( $$ ) {
    my ($val, $max) = @_;

    my @array = split ',',( "$val,"x$max . $val );
    return \@array;
}

# compute the sum of two arrays
sub add ( @@ ) {
    my ($a, $b) = @_;
    my @c = ();
    my $limit = min($#$a, $#$b);
    for($i=0; $i<=$limit; $i++) {
        $c[$i] = $a->[$i] + $b->[$i];
    }
    return \@c;
}

# compute the day-by-day (i.e. pointwise, array) product of two arrays
sub mult ( @@ ) {
    my ($a, $b) = @_;
    my @c = ();
    my $limit = min($#$a, $#$b);
    for($i=0; $i<=$limit; $i++) {
        $c[$i] = $a->[$i] * $b->[$i];
    }
    return \@c;
}

# compute the dot product (sum of day-by-day products) of two arrays
sub dot ( @@ ) {
    my ($a, $b) = @_;
    my $c = 0;
    my $limit = min($#$a, $#$b);
    for($i=0; $i<=$limit; $i++) {
        $c += $a->[$i] * $b->[$i];
    }
    return $c;
}


#--------------------fungicide setup and subroutines-----------------
# The following allows the infection rate to be high, then kills some
# infections off. Requires separate mechanisms for kickback, persistence.

# %templates is a hash of refs to coefficients and arrays of daily
# factors/coefficients.  keys are made from these tokens 'quad', 'quilt' or 'tilt' (fungicides),
# and 'inoculum', 'source', 'kickback' or 'persist'. 

# Infection is either 'exposed' or 'covered'(by sheath) and either latent
# or erumpent; erumpent pustules (may) emit spores from both inner and
# outer surfaces.  Accordingly, we have the cases 'lat', 'eru-in', and
# 'eru-out'.  The fungicide cases are 'none', 'quadris', 'tilt' and 'quilt'.

# For exposed latent infections (ie new "primaries"=aerial-spread), we use a table indexed 
# by day, and persistence matters.  Otherwise, only currently extant 
# infections are affected.

# hash value is a hash of survival rates indexed by day difference.
# for kickback, ->[0] is for today and ->[$i] is for $i days ago.
# For 'persist', ->[$i] is for $i days in the future, and ->[0] is unused.

BEGIN {

   # the fraction of the pustules which release spores on the inner
   # surface (supplying spores for contact-spread infections if pustules are on a sheath)
   $inner_frac = 0.9;

    # These hash keys and values are applied on the day of spray.  
    # Values are survival proportions.

    %factors = (
                'quadris/covered' =>    0.355,
                'quilt/covered' =>      0.222,
                'tilt/covered' =>       1,
                'none/covered' =>       1,


              # The eei cohorts have already been multiplied by 0.9 to create them;  
              # to get observed 3.1% for quadris (for example), correct for that here.  
                'quadris/exp/eru-in' => 0.031/$inner_frac,
                'quilt/exp/eru-in' =>   0.022/$inner_frac,
                'tilt/exp/eru-in' =>    0.7/$inner_frac,  
                'none/exp/eru-in' =>    $inner_frac/$inner_frac,

              # survival of ee pustules exposed to fungicide, from exptal data     
                'quadris/exp/eru-out' => 0.044,       
                'quilt/exp/eru-out' =>  0.044,
                'tilt/exp/eru-out' =>   0.56, 
                'none/exp/eru-out' =>   1,

              # to modify inoc'm prod'n in inoculum subroutine, from exptal data
                'quadris/exp/sporulation' => 0.018,  
                'quilt/exp/sporulation' =>  0.013,
                'tilt/exp/sporulation' =>  0.26,
                'none/exp/sporulation' =>   1

                );

}

BEGIN {
    # day_inf: day of infection
    # day_spr: day of spray
    # delta_day = day_inf-day_spray: >0 is persistence, <=0 is kickback
    # frac_{quadris,tilt}: proportion surviving treatment
    # NOTE!: model assumes incorrectly that newly-emerged plant area will have
    #  fungicide persisting on it. This may be partially true due to systemic nature, but not precisely

#delta_day  frac_quadris  frac_quilt  frac_tilt
    $table = <<'EOTABLE';
-15	0.044	0.044	0.560
-14	0.044	0.044	0.560
-13	0.040	0.040	0.520
-12	0.031	0.031	0.472
-11	0.027	0.027	0.400
-10	0.022	0.022	0.336
-9	0.018	0.018	0.280
-8	0.013	0.013	0.232
-7	0.013	0.013	0.184
-6	0.009	0.009	0.144
-5	0.009	0.004	0.112
-4	0.009	0.004	0.080
-3	0.013	0.004	0.056
-2	0.013	0.009	0.040
-1	0.018	0.009	0.032
0	0.018	0.009	0.032
1	0.022	0.009	0.032
2	0.031	0.009	0.040
3	0.036	0.013	0.048
4	0.044	0.013	0.072
5	0.053	0.018	0.096
6	0.062	0.027	0.128
7	0.071	0.040	0.160
8	0.080	0.058	0.208
9	0.093	0.071	0.256
10	0.107	0.093	0.312
11	0.120	0.120	0.368
12	0.133	0.133	0.440
13	0.151	0.151	0.512
14	0.169	0.169	0.592
15	0.182	0.182	0.672
16	0.222	0.222	0.800
17	0.267	0.267	0.800
18	0.311	0.311	0.800
19	0.356	0.356	0.800
20	0.400	0.400	0.800
21	0.444	0.444	0.800

EOTABLE

    # turn the string $table into a long array
    @rawtable = split '\s+', $table; 

    # 'map' is a loop, applying a function to everything in an array.
    # Make a hash with array references as values, names tilt/kickback etc as keys
    map { $templates{$_} = [] } qw(tilt/kickback tilt/persist
                                quadris/kickback quadris/persist
                                quilt/kickback quilt/persist
                                );

    #loop while there are elements in @rawtable, assign them to these variables 
    while(@rawtable) { 
        my $offset = shift @rawtable;
        my $quadris= shift @rawtable;
        my $quilt =  shift @rawtable;
        my $tilt   = shift @rawtable;

        # for each set of 4 in the while loop, test whether it's persistance
        #   or kb, then put the proportion in as a value for appropriate key
        #   in templates hashes
        if ($offset <= 0) { # kickback
            # spray is -$offset days after infection; 0th element is used
            $templates{'quadris/kickback'}->[-$offset] = $quadris;
            $templates{'quilt/kickback'}->[-$offset]   = $quilt;
            $templates{'tilt/kickback'}->[-$offset]    = $tilt;
        } else { # persistence
            # infection is $offset days before spray; 0th element is not used
            $templates{'quadris/persist'}->[$offset] = $quadris;
            $templates{'quilt/persist'}->[$offset]   = $quilt;
            $templates{'tilt/persist'}->[$offset]    = $tilt;


        }
    } #  closing bracket for 'while' 
} # closing bracket for 'begin'

sub persist_template( $$$ ) {
    # from fcide, sprayday and proportions, makes factor for each day of season
    my ($fcide_name, $spraydays, $factor) = @_;  
    my $templ = $templates{"$fcide_name/persist"};
    foreach $sprayday (@$spraydays) {
        for($i=1; $i<=$#$templ; ++$i) {
            my $day = $sprayday+$i;
            last if $day<0;
            $factor->[$day] *= $templ->[$i]; 
        }
    }
}

# arguments are references to @quadris_days and @tilt_days; return val
sub make_fungicide_factor ( $$$ ) {
    my ($quadris_days, $quilt_days, $tilt_days) = @_;
    # expresses persistence in the form of an array ref of daily survival rates
    #   for new latent exposed infections.
    my $persist = new_array(1, $daymax);
    persist_template( "quadris", $quadris_days, $persist);
    persist_template( "quilt", $quilt_days, $persist);
    persist_template( "tilt", $tilt_days, $persist);

    # identifies day(s) on which fungicide(s) are applied
    my %kb_events = ();
    foreach my $qd (@$quadris_days) {
        $kb_events{$qd} = "quadris";
    }
    foreach my $ud (@$quilt_days) {
        $kb_events{$ud} = "quilt";
    }
    foreach my $td (@$tilt_days) {
        $kb_events{$td} = "tilt";
    }

    return $persist, \%kb_events; #an array (?) and a hash ref
}

# DESPITE ITS NAME, the following implements ALL fungicide effects
# on the current population (i.e. everything except persistence).

sub apply_kickback ( $$ ) {
    my ($fcide, $cohorts) = @_;
    my @kickback = @{ $templates{"$fcide/kickback"} };

    foreach my $cohort (@$cohorts) {
        if ($cohort->{'le'}>0) {
            # latent (le) killed by fungicide disappear per fungicide Table, 
            # as if they never existed  
            if ($cohort->{'infday'} + $#kickback >= $today) {
                my $age =  $today-$cohort->{'infday'};
                $cohort->{'le'} = $cohort->{'le'}*$kickback[$age];
            }
        }
 
        if ($cohort->{'lc'}>0) {
            # some covered latent pustules die also, as per %factors hash
            my $factor = $factors{"$fcide/covered"};
            $cohort->{'lc'} = $cohort->{'lc'}*$factor;

        }

        if (exists $cohort->{'eruday'}
            and $cohort->{'eruday'} <= $today) {
            # some erumpunt (ee) pustules, and different propn of ec pustules, die 
            my $eelost = $cohort->{'ee'} * (1 - $factors{"$fcide/exp/eru-out"}); 
            my $eclost = $cohort->{'ec'} * (1 - $factors{"$fcide/covered"}); 


            $cohort->{'de'} += $eelost;
	    $cohort->{'dc'} += $eclost;
            $cohort->{'ee'} -= $eelost;
            $cohort->{'ec'} -= $eclost;

            # some eei (source for sec infs) die as per %factors hash 
            $cohort->{'eei'} *= $factors{"$fcide/exp/eru-in"}; 
            $cohort->{'eci'} *= $factors{"$fcide/covered"};    

            # some pustules produce fewer spores. The prop'n still produced is  
            #  calculated for use by sub inoculum
            $cohort->{'fungspor'} *= $factors{"$fcide/exp/sporulation"};          
        }
    }
}

#Action thresholds for fungicides
sub thresholdTriaz ( $ ) {             #action threshold for triazole fungicide in decision aid
    my ($day) = @_;
    return 75 if $day < 128;
    return 4000 if $day > 171;
    return (10**((0.04*$day) - 3.23));

    return $thresholdTriaz;
}

sub thresholdStrob ( $ ) {             #action threshold for strobilurin fungicide in decision aid
    my ($day) = @_;
    return 500 if $day < 99;
    return (10**((0.0022*$day) + 3.34)) if $day > 133;
    return (10**((0.026*$day) + 0.156));
    return $thresholdStrob;
}

sub plant_area ( $ ) {     # empirically-derived polynomial for plant area over LUP
    my ($day) = @_;
    my $lup = $lup[$day];
    return 976 if $lup < 0.588;
    return 8858 if $lup >= 8.9;
    return polynomial($lup, 1786, -2919, 2911, -505.1, 25.27);
}

# snum is the sheath number sheath_a, sheath_b, sheath_c and sheath_d defined above
sub sheath_length ( $$ ) {
    my ($snum, $day) = @_;
    if ($snum<1 or $snum>4) {  
        ($packagename, $filename, $line) = caller;
        warn "bad sheath number $snum; caller ($packagename, $filename, $line)";
    }
    my $lup = $lup[$day];
    my ($a, $b, $c, $d) = ($sheath_a[$snum], $sheath_b[$snum], $sheath_c[$snum], $sheath_d[$snum]) ;
    my $sheath_length = (($a/(1+exp(($b-$lup)/$c))) - ($d*$a)) unless $c <= 0;
    if ($sheath_length < 0) {
       $sheath_length = 0;
    }
    return $sheath_length;
}

# snum is the stem number stem_a, stem_b, stem_c and stem_d as for sheaths
sub stem_length ( $$ ) {
    my ($snum, $day) = @_;
    if ($snum < 0 or $snum > 4) {
        ($packagename, $filename, $line) = caller;
        warn "bad stem number $snum; caller ($packagename, $filename, $line)";
    }
    my $stem_length;
    my $lup = $lup[$day];
    my ($a, $b, $c, $d) = ($stem_a[$snum], $stem_b[$snum], $stem_c[$snum], $stem_d[$snum]);

    my $stem_length = (($a/(1+exp(($b-$lup)/$c))) - ($d*$a)) unless $c <= 0;
    if ($stem_length < 0) {
       $stem_length = 0;
    }
    return 0 if ($day < $stemstart);  # no growth yet by definition
    return $stem_length;
}

# barestem is the exposed stem below the node, ie non-sheath 'stem'.
sub barestem_length ( $$ ) { 
    my ($snum, $day) = @_;
    my $sheath_length = sheath_length($snum, $day);
    my $stem_length = stem_length($snum, $day);
    my ($a, $d) = ($sheath_a[$snum], $sheath_d[$snum]) ; 
    my $sheathmax = $a*(1-$d);
    my $barestem_length;
    if ($sheath_length >= 0.85*$sheathmax) {
        $barestem_length = $stem_length - $sheath_length;
        if ($barestem_length<= 0){
            $barestem_length = 0;
        }
    } else {
        $barestem_length = 0;
    }
    return $barestem_length;
}

sub get_parts_area ( ) {  #calculate areas of st0, sheaths, barestems and 'other'
    #places: 0=st0, 1-4=sh1-4, 5-8=barestem1-4, 9=other (the sheath-covered stem not needed here)
    @parts_areas=(0,0,0,0,0,0,0,0,0,0); 
    my $day = $today;
    my $plant_area = plant_area($today);
    my $snum = shift @_;
    my $bsnum = $snum + 4;

    # the following calculates projected area (ie 0.5*circumference), uses 1.7mm as diameter
    foreach $snum (1..4) {    
       $parts_area[$snum] = &stem_length($snum,$day) * 3.14 * 0.5 * 0.17 * $tillers;        
       $parts_area[$snum + 4] = &barestem_length($snum,$day) * 3.14 * 0.5 * 0.17 * $tillers;
    }

    $parts_area[0] = &stem_length(0,$day) * 3.14 * 0.5 * 0.17 * 1.2 * $tillers; # larger, due to inflorescence            
    $all_stems_area = @parts_area[0] + @parts_area[1] + @parts_area[2] + @parts_area[3] 
        + @parts_area[4] + @parts_area[5] + @parts_area[6] + @parts_area[7] + @parts_area[8];
    $parts_area[9] = $plant_area - $all_stems_area;
    $parts_area[9] <= 0 ? 0 : $parts_area[9];  #don't let it be negative
    #store the value of 'other' area, by day, for later use
    $area_oth_stored[$day] = $parts_area[9];   

    return @parts_area;  
}


# elongation: snum is the stem number;
sub stem_growth ( $$ ) {
    my($snum,$day) = @_;
    if ($snum<0 or $snum>3) {
        ($packagename, $filename, $line) = caller;
        die "bad stem number $snum; caller ($packagename, $filename, $line)";
    }
    return 0 if ($day < $stemstart);  # no growth yet by definition
    my $stem_growth = &stem_length($snum,$day) - &stem_length($snum,$day-1);
    if ($stem_growth < 0) {
       $stem_growth = 0;
    }
    return $stem_growth;
}


sub new_secondaries (){        # 'secondaries' = contact-spread infections
    # how-to question here.  We have information about where on the sheath
    # the source infections are.  But if we use it, infection density isn't
    # a constant function of length and we would end up with a
    # rather messy set of cohorts - a 2-d array of infection density as a 
    # function of space and time.  Instead, model pretends that the infections
    # are uniformly distributed along the sheath.

    # calculate the number of source (for secondaries) infections on each sheath
    @src_infs=(0,0,0,0,0);  # first 0 is a placeholder
    foreach $c (@cohorts) {
        if ($c->{'on'} eq 'sheath' and exists $c->{'eruday'}) {
            $src_infs[$c->{'snum'}] += $c->{'eei'} + $c->{'eci'};
        }
    }

    $opty_src_infs=\@src_infs;  #this is a ref to the src_infs array, for opt_y output
    my @infections = ();
    foreach $sh (1..4) {
        push @infections, (stem_infs($sh, $src_infs[$sh]));
    }
    return @infections;
}

#the sub new_secondaries (above) creates an array called infections, which is 
#built by using the subroutine stem_infs (below) that 
#needs the source infs calculated within sub new_secondaries

# ssh is the source sheath; srcs is the number of spore-producing source
# pustules on that sheath per modeled area (30 cm row), including fungicide factors.
# This is a mishmash of sheath->sheath and sheath->stem infections.  The
# simplifying assumption is: what's under sheath n is sheath n-1 or barestem,
# as long as n>1.  If n==1, what's under it is the flower stem (st0).

sub stem_infs( $$ ) {
    my ($ssh, $srcs) = @_;
    warn "bug: bad sheath number $ssh" if ($ssh < 1 or $ssh > 4);
    my $dsh = $ssh-1;  # destination and source sheath numbers
    my $mvmt_today = stem_growth($dsh,$today);
    $mvmt_today = 0 if $mvmt_today<0;
 
    # don't allow new secondaries to exceed the space available each day
    #  (i.e., a cap on overlaps).
    if ($ssh>1 ) {
       $infs = $srcs * $stem_infs_pcmL * $mvmt_today; 

       if ($infs > ($tillers * $mvmt_today * ($stem_infs_pcmL))) {
           $infs = ($tillers * $mvmt_today * ($stem_infs_pcmL))

       } 
     }

     if ($ssh==1 ) { #this is for infs on stem0, which has a different area/length, due to inflorescence 

       $infs = $srcs * 1.2 * $stem_infs_pcmL * $mvmt_today;

       if ($infs > ($tillers * $mvmt_today * ( 1.2 * $stem_infs_pcmL))) {
           $infs = ($tillers * $mvmt_today * ( 1.2 * $stem_infs_pcmL))

       } 
     }

#  following is because stem eqns may produce stem longer than sheath too early, ie before the stem actually emerges from enclosing sheath
    if ($dsh == 0) {
	   $on = 'stem' 
    }

     elsif (&barestem_length($dsh, $day) > 0) { 
	     $on = 'barestem' 
     }
	   else { 
	     $on = 'sheath' 
	   }

    return &new_cohort("sheath", $on, $infs, $ssh, $dsh);
}

# make names for the plant parts used in arrays
#places: 0=st0, 1-4=sh1-4, 5-8=barestem1-4, 9=other
BEGIN {
@plant_parts=();
$plant_parts[0]="stem0";
$plant_parts[1]="sheath1";
$plant_parts[2]="sheath2";
$plant_parts[3]="sheath3";
$plant_parts[4]="sheath4";
$plant_parts[5]="barestem1";
$plant_parts[6]="barestem2";
$plant_parts[7]="barestem3";
$plant_parts[8]="barestem4";
$plant_parts[9]="other";       # "other" obtained by subtracting modeled sheath&stem areas from "plant_area"
}

sub get_cohort_plant_part ( $ ) { # which plant part is this cohort on? 
    my $c = shift @_;
    my $ppart;
    if ($c->{'on'} eq 'sheath' ) {
	$ppart = "sheath" . $c->{'snum'};
    } elsif ($c->{'on'} eq 'barestem' and $c->{'snum'} !=0) {
	$ppart = "barestem" . $c->{'snum'};
    } elsif ($c->{'on'} eq 'stem' and $c->{'snum'}==0 ) {
	$ppart = "stem0";
    } elsif ($c->{'on'} eq 'other') { 
	$ppart = "other";
    } else {
	warn "confusing cohort location on=" . $c->{'on'};  #catch logic errors
    }
    return $ppart;
}

# need to account for area unavailable to new aerial-spread infection because it is already-occupied
# count occupied 'exposed'sites ('covered' ones don't matter for limiting aerial-spread infs, because 
#    that area is inaccessible except by contact-spread)
sub count_filled_sites ( ) {  
    my %filled_sites=();
    for $plant_part (@plant_parts) {
	$filled_sites{$plant_part}=0;
    }
    
    foreach $c (@cohorts, @dead_cohorts) { 
	my $ppart = get_cohort_plant_part( $c );
	$filled_sites{$ppart} += $c->{'de'} + $c->{'ee'} + $c->{'le'};
    }
    return \%filled_sites;
}

sub count_sec_infs ( ) {  #this is only used for illustration or debugging output, not part of model operation
   @sec_infs=(0,0,0,0,0); # for stems 0..4
   foreach $c (@cohorts) {
      foreach $snum (0..4) { 
        if ($c->{'from'} eq 'sheath'and $c->{'snum'} == $snum) {
    #      all of them except dead covered
    #      $sec_infs[$snum] +=  $c->{'lc'} + $c->{'le'} + $c->{'ec'} + $c->{'ee'} + $c->{'de'} ; }

    #    if want just the exposed ones:
         $sec_infs[$snum] +=  $c->{'ee'} + $c->{'le'} + $c->{'de'}; }

    #    if want just the erumpent exposed ones                             
    #    $sec_infs[$snum] +=  $c->{'ee'} ; }                                                      
      }
    }
    return @sec_infs;
}  # end bracket for sub count_sec_infs


#  the following lines are to create values for output of all exposed erumpent  
#  'secondaries' (contact-spread infections). Created only for output, not part of model operation.
sub count_sec_vis ( ) {  
   @sec_vis=(0,0,0); # for stems 0..3. There are no secs on st4
   foreach $c (@cohorts) {
      foreach $snum (0..3) { 
        if ($c->{'from'} eq 'sheath'and $c->{'snum'} == $snum) {
                                     
        $sec_vis[$snum] +=  $c->{'ee'} ; }  # just the erumpent exposed ones                                                      
      }
    }
    return @sec_vis;
} # end sub count_sec_vis


# a cohort is a set of infections with same age, location, and mechanism.
# cohorts are represented as hash references with values for these keys:

# infday, eruday: days of year of infection and eruption;
# lup:          lup at infection initiation
# on:           Where the infection is: 'stem', 'sheath', 'barestem' or 'other'.
#               prim'y on sheath ->'sheath', prim'y elsewhere ->'other' or 'barestem'.
# snum:         The sheath or stem number the inf is on, or undefined
# from:         The source/type of the infection 'primary' or 'sheath' ( == 'aerial-spread' or 'contact-spread')
# srcnum:       if "from" = 'sheath', this is the sheath number, else undefined
# serial no:    A unique identifier for each cohort, for debugging
# cover:        the distance, in cm below stem/sheath top, of the auricle of
#               the next sheath down.  undef for on=other
# 'fungspor':   relative sporulation proportion for fungicide-affected ee's

# le, lc, ee, ec, eei, eci, d: number of members of the cohort per modeled field area (900 cm2) in
# life states "latent exposed", "latent covered", "erumpent exposed",
# "erumpent covered", "erumpent exposed (inner)", "erumpent covered
# (inner)", and "dead", respectively.  
# Paths- primary (aerial-spread) infections: le->ee->d. 
#        secondary (contact-spread) infections: lc->(ec or le)->ee->d.  
#        also can -> dead with fungicide.

#  eei and eci are not life stages per se -- an erumpent pustule on a
#  sheath releases spores on both the inside and the outside faces of the
#  sheath.  inside -> contact-spread infections; outside -> aerial (for 'primary infections')
#  Because fungicides affect release to the inside and outside
#  differently, it is convenient to think of them as separate entities.
#  Thus, when a latent pustule erupts, it is transformed into
#  two spore generators.  They are tracked separately, with the outer ones
#  tracked with keys ee and ec, and the inner ones with keys eei and eci.
#  Only the outer ones are counted for disease severity bookkeeping.

BEGIN { $cohort_serial = 2; }

# from: inoculum source: 'primary', or 'sheath n' for secondary (contact-spread).
# (from, on, srcnum, snum) may be: 'primary', 'sheath', undef, n;
# 'primary', 'other', undef, undef; 'sheath', 'stem', n, n-1; or 'sheath',
# 'sheath', n, n-1.

sub new_cohort ( $$$;$$ ) {
    my ($from, $on, $size, $srcnum, $snum) = @_;

    if ($stderrout) { 
        my @x = ("new_cohort", $today, $from, $on, $size, $srcnum, $snum);
        @x = map { $_= defined $_ ? $_ : 'undef' } @x;
    }

    if ($size < $MIN_COHORT_SIZE) { # too small; ignore
        return ();
    }

    my %c = ();
    $c{'serial no'} = $cohort_serial++;
    $c{'infday'} = $today;
    $c{'lup'} = $lup[$day];
    $c{'from'} = $from;
    $c{'on'} = $on;
    $c{'le'} = 0;
    $c{'lc'} = 0;
    $c{'ee'} = 0;
    $c{'ec'} = 0;
    $c{'eei'} = 0;
    $c{'eci'} = 0;
    $c{'de'} = 0;
    $c{'dc'} = 0;
    $c{'exp frac'} = 1;
    $c{'fungspor'} = 1;  # factor (0-1) for fungicide effect on sporulation

    if ($from =~ /sheath/) { # need to know location for exposure, and for contact-spread infection
        ($packagename, $filename, $line) = caller;
        die "bad snum ($snum); caller ($packagename, $filename, $line)"
            if ($snum < 0 or $snum > 3);
        die "bad srcnum ($srcnum); caller ($packagename, $filename, $line)"
            if ($srcnum < 1 or $srcnum > 4);
        $c{'snum'} = $snum;
        $c{'init_stem_len'} = stem_length($snum, $today);
        $c{'init_sheath_len'} = sheath_length($srcnum, $today);
        $c{'exp frac'} = 0;   # contact-spread inf cohorts start totally covered
    } else {
        $c{'snum'} = $snum if defined $snum;
    }

    if ($from =~ /sheath/) { # contact-spread infection mechanism
        $c{'srcnum'} = $srcnum;
        # infection starts latent and covered
        $c{'lc'} = $size;
    } else {
        # 'primary' (aerial-spread) infection starts latent and exposed
        $c{'le'} = $size;
        $c{'srcnum'} = 0;  # bookkeeping convenience
    }
    return (\%c);
}

sub age_cohort( $ ){
    my $c = shift @_;  # $c is a hash reference when this sub is called
    my $lup = $lup[$day];  
    if (exists $c->{'eruday'}) {          # if it was already erumpent   
          if ($lup >= $c->{'lup'} + $pust_life) { 
            # dies of old age
            for $stage (qw( le lc ee ec )) { 
                $c->{'de'} += $c->{'ee'};
                $c->{'dc'} += $c->{'ec'};
                $c->{'ee'} = 0;
                $c->{'ec'} = 0;
                $c->{'eei'} = 0;
                $c->{'eci'} = 0;
            }
          return 0;  # return value of 0 for this sub means the cohort is not alive 
        }
    } elsif ($lup >= $c->{'lup'} + 1 ) {  # time to erupt
        $c->{'ee'}  = $c->{'le'};
        $c->{'ec'}  = $c->{'lc'};
        if ( $c->{'on'} eq 'sheath'){
           $c->{'eci'} = $c->{'lc'} * $inner_frac;
           $c->{'eei'} = $c->{'le'} * $inner_frac;
        }
        $c->{'lc'}  = 0;
        $c->{'le'}  = 0;
        $c->{'eruday'} = $day;
    }

    my $latent = $c->{'le'}+$c->{'lc'};
    #are there still live ones in this cohort?
    my $alive = $latent+$c->{'ee'}+$c->{'ec'} >0 ? 1 : 0;  

    return $alive;
}


# sheath+stem growth; move some lc to le and some ec to ee; some pustules covered by sheath become exposed as the internode elongates
sub expose_cohort ( $ ) {
    my $c = shift @_;
    if ($c->{'from'} eq 'sheath') {

        if (not defined $c->{'exp frac'}) {
            dumphash('no exp frac', $c);
            croak();
        }
        my $old_exp_frac = $c->{'exp frac'};

        #no need to calculate if it's already completely exposed
        return if $old_exp_frac == 1;         
       
        # estimate exposed portion
        my $expfrac = 
            (stem_length($c->{'snum'}, $today)- $c->{'init_stem_len'})
                /$c->{'init_sheath_len'};

        # force to [0,1] for sanity
        $expfrac = 1 if $expfrac > 1;
        $expfrac = 0 if $expfrac < 0;
#       portion of previously covered part of the cohort which is now exposed.
        my $portion = ($expfrac - $old_exp_frac)/(1-$old_exp_frac);

        # for each stage, expose portion of that previously covered
        my $pop_revealed = $c->{'lc'} * $portion;
        $c->{'le'} += $pop_revealed;
        $c->{'lc'} -= $pop_revealed;

        $pop_revealed = $c->{'ec'} * $portion;
        $c->{'ee'} += $pop_revealed;
        $c->{'ec'} -= $pop_revealed;

        $c->{'exp frac'} = $expfrac;
    }
}

sub inoculum() {   
    my($inoculum,$cohort);
    $inoculum = 0;
    $inoc_initial = 0;
    foreach $cohort (@cohorts) {
        if ($cohort->{'ee'} > 0) {
            my $days_old = $day - $cohort->{'eruday'};
            $inoc_initial += $cohort->{'ee'}*$inoc_schedule[$days_old]; 
            $inoculum += $cohort->{'ee'}*$inoc_schedule[$days_old]*$cohort->{'fungspor'};
#                previous line includes fungicide factor to reduce sporulation if $sprayday is after $eruday
        }
    }
    $inoculum->[$day] = $inoculum;
    $inoculum *= power(10, -$rain[$today]/$prior_rain_constant); #heavy rain reduces inoculum

    return $inoculum;
}


# susceptibility declines as plant part ages
# here is a helper function for use in susceptibility();
sub susceptility_math( $$$ ) {
    my ($a, $b, $c) = @_;
    my $susc;

    if ($lup < $a) { $susc=1 }
    elsif ($lup > $b) { $susc=(1 - $c*($b - $a)) }
    else { $susc = 1  - $c*($lup-$a) }

    return $susc;
}

# return susceptibilities of plant parts; 
# estimate age-related changes in susceptibility for each component (decreases 
#   starting at full extension [defined as 1 cm short of asymptotic max], goes to 0.01X )                  
sub susceptibility() {      
    my %susc = ();

    $susc{"stem0"} = susceptility_math( 8.5, 10.0, 0.66 );
    $susc{"sheath1"  } = susceptility_math( 7.5, 9.0, 0.66 );
    $susc{"sheath2"  } = susceptility_math( 6.5, 8.0, 0.66 );
    $susc{"sheath3"  } = susceptility_math( 5.0, 6.5, 0.66 );
    $susc{"sheath4"  } = susceptility_math( 4.0, 5.0, 0.99 );
    $susc{"barestem1"} = $susc{"sheath1"};
    $susc{"barestem2"} = $susc{"sheath2"};
    $susc{"barestem3"} = $susc{"sheath3"};
    $susc{"barestem4"} = $susc{"sheath4"};
    $susc{"other"    } = susceptility_math( 2.7, 7.0 , 0.23 );
    return \%susc;
}


sub new_primaries() {    # return list of new primary-infection (ie aerial-spread) cohorts
    my $plant_area = plant_area($today);
    my %new_infs;
    my $inoculum = inoculum();   
    my $lup = $lup[$day];
    my $inf_factor = $inf_factor[$today];

#   figure available sites (uninfected area) from plant_area: total infections so far, for each plant part

    my $filled_sites = &count_filled_sites();  # hash ref

    my @parts_area = &get_parts_area();
    # translate from numeric locations to named locations
    my %parts_area_hash = ();
    for $i (0..9) {
	$parts_area_hash{$plant_parts[$i]} = $parts_area[$i];
    }

    my %avail = ();
    for $ppart (@plant_parts) {
	$avail{$ppart} = $pustules_per_cm2 * $parts_area_hash{$ppart} - $filled_sites->{$ppart}; 
    }   

    my $susc=susceptibility();


# Here is the basic (verbal) form for the aerial-spread infection equation, which is made more compact 
# (and less understandable) further on:
#    new aerial-spread infections = 
#       $susc*fungicide$persistence*uninfected_area*(1-exp(-$k_infection*$inoculum*$inf_factor /$plant_area));

#   Fungicide component of eqn could be modified (w lots of add'l code) to model lack of fungicide 
#   on area emerging post-application, but this is not in the current version of the model

#  following lines cause $k_infection to increase as function of plant size/development (correlated with LUP); 

   my $lup = $lup[$day];   
   my $LUP_kmin = 4.0; 
   my $LUP_kmax = 5.2; 
   my $kMIN = 40/1000;      
    if ($lup < $LUP_kmin) { $k_infection = $kMIN }
    elsif ($lup > $LUP_kmax) { $k_infection = $kMAX }
    else { $k_infection = (($kMAX - $kMIN) / ($LUP_kmax - $LUP_kmin)) * ($lup - $LUP_kmin) + $kMIN}

#  next two lines combine some terms to make the equations for aerial-spread infections of each part more compact 
    $kInocInf = $k_infection*$inoculum*$inf_factor;
    my $all_sites = $pustules_per_cm2 * $plant_area;

# the following is the compact version of the aerial-spread infection equation
    for $p (@plant_parts) {
	if ($avail{$p} <= 0) { $new_infs{$p} = 0; }
	else {
        $new_infs{$p} = $susc->{$p} * $persistence->[$today] * $avail{$p}*( 1 - exp( - $kInocInf/$all_sites));
	}
    }     # end bracket for aerial-spread infection equations on the plant parts

#  the following lines assign the results of the aerial-spread infection equations
    my @ret = ();
    for my $shth (1..4) {
        push @ret, new_cohort('primary', 'sheath', $new_infs{"sheath".$shth},
                              undef, $shth);
    }

    for my $brstm (1..4) {
        push @ret, new_cohort('primary', 'barestem',
			      $new_infs{"barestem".$brstm}, undef, $brstm);
    }
    push @ret, new_cohort('primary', 'stem',
			      $new_infs{"stem0"}, undef, 0);
    push @ret, new_cohort('primary','other', $new_infs{"other"}, undef, undef);

    $opty_susc = $susc;
    $opty_parts_area = \%parts_area_hash; 
    $opty_avail = \%avail;
    $opty_filled = $filled_sites;
    $opty_new_infs = \%new_infs;

    return @ret;
}

# following sub is to allow disease (latent and erumpent) per modeled field area to decrease due to lost plant area 
# (from senescence). Addresses 'other'area only, not whole-plant area, as it's the leaves and lower-internode sheaths that die 

sub plant_area_decrease() {  
      
  return unless defined $lup_ago;  # can't do this at start of data
      # How big is the 'other' area, relative to its size yesterday?

  $tdy = $today; 

  if ($today < 120) { 
     $size_ratio = 1 
  }
  else { $size_ratio = $area_oth_stored[$tdy]/$area_oth_stored[$tdy - 1] 
  }

  if ($size_ratio > 1) { $size_ratio = 1 }

# If plant has shrunk, kill off a percentage of each 'other' cohort.
  if ($size_ratio < 1) { 
        foreach my $cohort (@cohorts) {
            if($cohort->{'on'} eq 'other') { 
                foreach my $stage (qw(lc le ec ee de dc)) {

                    $cohort->{$stage} *= $size_ratio;
                 }
             }

             #for possible model modification, if one wants to apply decrease to sh,st,barest plant parts
             else {        
                   #this is where I can adjust coefficients
                   my $new_ratio = ($size_ratio*.20) + (1.00 * .80);
 
                   #default (NO loss of sh,st,bst pustules w decreasing plant area)
 		   $new_ratio = 1.0;  

                foreach my $stage (qw(lc le ec ee de dc)) {
                   $cohort->{$stage} *= $new_ratio;
                }
             } # end else
         }     # end foreach my $cohort
  }            # end if size ratio
}              # end sub decrease

# following is to calculate the area under disease progress curve for yield-critical window, set here to days 160-180
# this code uses the (alive + recently dead) disease; could choose instead to include only the living pustules in 'disease'
# This is not part of model operataion, is just for output if wanted
 
sub AUDPC ( $ )  {
if ($day < 160 or $day >180)  {
    $AUDPC = 0;
  }  else   {
    $AUDPC = $AUDPC + $vis_AandRecD;
#    $AUDPC = $AUDPC + $vis_live;    #un-comment this line (and comment-out previous) to get living-pustule disease only
  }
}

# set up the output -- open the file, select it, and set the formats
# if the argument is undefined, use STDOUT
sub set_output( $ ) {
    my $outfile = shift @_;

    if (defined $outfile) {
        open OUTFILE, ">$outfile"
            or die "can't open temporary output file $outfile";
        select OUTFILE;
    } else {
        select STDOUT;
    }

    print "$header_string\n";   #first page only
    $FORMAT_NAME     = "sim_out";
    if ($opt_x) {
        # Output will be read as text; include page headers
        $FORMAT_TOP_NAME = "sim_out_TOP";
        $FORMAT_LINES_PER_PAGE = 23;
    } else {
        # Non-interactive -- suppress page headers
        $FORMAT_TOP_NAME = "null";
        $FORMAT_LINES_PER_PAGE = 400;
    }
}

sub stop_output() {
    close(select(STDOUT)) unless select() =~ /STDOUT$/;
}

print STDERR "========= $0 end =========\n";
